/*
 * Copyright (C) Cyril Hrubis <chrubis@suse.cz>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/*
 * Measures pselect timeouts and writes them to a file for further analysis.
 */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#define SIZE 1000

static struct timespec start_time, stop_time;
static clockid_t clock_id;

void tst_timer_start(clockid_t clk_id)
{
	clock_id = clk_id;

	if (clock_gettime(clock_id, &start_time))
		perror("clock_gettime()");
}

void tst_timer_stop(void)
{
	if (clock_gettime(clock_id, &stop_time))
		perror("clock_gettime()");
}

static inline struct timespec tst_timespec_diff(struct timespec t1,
                                                struct timespec t2)
{
	struct timespec res;

	res.tv_sec = t1.tv_sec - t2.tv_sec;

	if (t1.tv_nsec < t2.tv_nsec) {
		res.tv_sec--;
		res.tv_nsec = 1000000000 - (t2.tv_nsec - t1.tv_nsec);
	} else {
		res.tv_nsec = t1.tv_nsec - t2.tv_nsec;
	}

	return res;
}

struct timespec tst_timer_elapsed(void)
{
	return tst_timespec_diff(stop_time, start_time);
}

static inline long long tst_timespec_to_us(struct timespec t)
{
	return t.tv_sec * 1000000 + (t.tv_nsec + 500) / 1000;
}

static inline long long tst_timer_elapsed_us(void)
{
	return tst_timespec_to_us(tst_timer_elapsed());
}

static int cmp(const void *a, const void *b)
{
	const long long *aa = a, *bb = b;

	return *aa > *bb;
}

static void test_pselect(unsigned int sleep)
{
	long long m[SIZE];
	long long mean = 0;
	int i;
	fd_set readfds;
	char fname[256];
	FILE *f;

	for (i = 0; i < SIZE; i++) {
		struct timespec tv;
		tv.tv_sec = sleep / 1000000000;
		tv.tv_nsec = sleep % 1000000000;

		FD_ZERO(&readfds);
		FD_SET(0, &readfds);

		tst_timer_start(CLOCK_MONOTONIC);
		pselect(0, &readfds, NULL, NULL, &tv, NULL);
		tst_timer_stop();

		m[i] = tst_timer_elapsed_us();
		mean += m[i];
	}

	qsort(m, SIZE, sizeof(m[i]), cmp);

	fprintf(stderr, "min %lli max %lli median %lli mean %.2f 0.1%% %.2f\n",
                m[0], m[SIZE-1], m[SIZE/2], 1.00 * mean/SIZE, 1.00 * sleep / 1000000);

	snprintf(fname, sizeof(fname), "pselect_timeout_%u.txt", sleep/1000);
	f = fopen(fname, "w");
	if (!f) {
		fprintf(stderr, "Failed to open %s\n", fname);
		return;
	}

	for (i = 0; i < SIZE; i++)
		fprintf(f, "%lli\n", m[i]);
	fclose(f);
}

int main(void)
{
	test_pselect(100000);
	test_pselect(500000);
	test_pselect(1000000);
	test_pselect(2000000);
	test_pselect(3000000);
	test_pselect(4000000);
	test_pselect(10000000);
	test_pselect(100000000);
	test_pselect(500000000);
	test_pselect(1000000000);

	return 0;
}
